#include "csprite.h"



// Initialize sprite.
CSprite::CSprite(int xTexture, int yTexture, int widthTexture, int heightTexture)
{
	m_xTexture = xTexture;
	m_yTexture = yTexture;
	m_widthTexture = widthTexture;
	m_heightTexture = heightTexture;
	m_xPos = 0;
	m_yPos = 0;
	m_xTexturePos = 0;
	m_yTexturePos = 0;
	m_isVisible = true;
}

// Clean up.
CSprite::~CSprite()
{
}

// Set Position.
void CSprite::setPos(int xPos, int yPos) {
	m_xPos = xPos;
	m_yPos = yPos;
}

// Move sprite x, y positions.
void CSprite::movePos(int xPos, int yPos) {
	m_xPos += xPos;
	m_yPos += yPos;
}

// Which part of the texture to use.
void CSprite::setTexturePos(int xTexturePos, int yTexturePos) {
	m_xTexturePos = xTexturePos;
	m_yTexturePos = yTexturePos;
}

// Set sprite visibility.
void CSprite::setVisible(bool isVisible) {
	m_isVisible = isVisible;
}

// Is the sprite visible ???
bool CSprite::getVisible(void) {
	return m_isVisible;
}

// Get X position
int CSprite::getXPos(void) {
	return m_xPos;
}

// Get Y position
int CSprite::getYPos(void) {
	return m_yPos;
}

// Get Width
int CSprite::getWidth(void) {
	return m_widthTexture;
}

// Get Height
int CSprite::getHeight(void) {
	return m_heightTexture;
}

// Collision test, only a simple inside squares test.
bool CSprite::inside(CSprite inSprite) {

	// Dont test if not visible.
	if (!m_isVisible) {
		return false;
	}

	// Here we go...
    if (((m_xPos + m_widthTexture) < inSprite.getXPos()) || (m_xPos > (inSprite.getXPos() + inSprite.getWidth())) ||
    ((m_yPos + m_heightTexture) < inSprite.getYPos()) || (m_yPos > (inSprite.getYPos() + inSprite.getHeight()))) {

    return false;
    } else {
    return true;
	}
}

// Render sprite.
void CSprite::addSpriteToGIFTag(char *& pMem, int Zvalue) {

	// Only render visible sprites.
	if (!m_isVisible) {
		return;
	}

	// pGIFReg is now pointing to the first memory position.
  	sps2GIFPackedRegister_t * pGIFReg = (sps2GIFPackedRegister_t *) pMem;

   	// Prepare the RGBAQ register
   	pGIFReg->i128=0;
   	pGIFReg->RGBAQ.s.R=0x80;	// Red
   	pGIFReg->RGBAQ.s.G=0x80;	// Green
   	pGIFReg->RGBAQ.s.B=0x80;	// Blue
   	pGIFReg->RGBAQ.s.A=0x80;	// Alpha
	pGIFReg++;

   	// Prepare the first UV register
   	pGIFReg->i128=0;
   	pGIFReg->UV.s.U=(m_xTexture + (m_xTexturePos*m_widthTexture)) << 4;
   	pGIFReg->UV.s.V=(m_yTexture + (m_yTexturePos*m_heightTexture)) << 4;
	pGIFReg++;

   	// Prepare the first XYZ2 register
   	pGIFReg->i128=0;
   	pGIFReg->XYZ2.s.X=(m_xPos + 1728) << 4;
   	pGIFReg->XYZ2.s.Y=(m_yPos + 1808) << 4;
   	pGIFReg->XYZ2.s.Z=Zvalue;
	pGIFReg++;

   	// Prepare the second UV register
   	pGIFReg->i128=0;
   	pGIFReg->UV.s.U=(m_xTexture + m_widthTexture + (m_xTexturePos*m_widthTexture)) << 4;
   	pGIFReg->UV.s.V=(m_yTexture + m_heightTexture + (m_yTexturePos*m_heightTexture)) << 4;
	pGIFReg++;

   	// Prepare the second XYZ2 register
   	pGIFReg->i128=0;
   	pGIFReg->XYZ2.s.X=(m_xPos + m_widthTexture + 1728) << 4;
   	pGIFReg->XYZ2.s.Y=(m_yPos + m_heightTexture + 1808) << 4;
   	pGIFReg->XYZ2.s.Z=Zvalue;
	pGIFReg++;

	// Update memory pointer.
	pMem = (char *)pGIFReg;
}