#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include <assert.h>
#include "PS2Defines.h"
#include "AddRegisters.h"
#include "ctexture.h"
#include "csprite.h"
#include "pad.h"

/*
Fourth example based on the tutorial from Henry S Fortuna...
A Space Invaders clone that uses sprites.

Read the seventeen first lessons from Henry S Fortuna. (http://www.hsfortuna.pwp.blueyonder.co.uk/)
The last lesson you read should be the Rendering Text With In-Built SPS2 Functions lesson.

This is pretty much the simplest sprites you can use, square sprites that represents a square from a bitmap texture...
We don't rotate or scale the sprites at all, so the current sprite class is pretty simple...

The game is built up by both static and dynamic sprites, by static sprites I mean sprites that are always rendered.
If a static sprite is not used, it is still rendered off screen.
The dynamic sprites are those defined in the sprite arrays.
I count the sprites before rendering them and only rendering the visible ones.

Hopefully you can figure out how the games is built by looking at the code, if not please E-mail me
and I'll be happy to try and explain it for you...

Have fun...

Written by Bamse 2004...
bamse@playstation2-linux.com
*/

// Function Prototypes
void AddGIFTag(char *& pMem, int numSprites);

// Our different game modes
void GameMenu(void);
void GamePlaying(void);
void GameOver(void);

// Called when switching mode in game.
void GameInit(void);
void GameOverInit(void);
void MenuInit(void);

// Work out how we are going to allocate SPS2 to each use:
// Here we have 4K allocated for primitives, that is right at the start of the memory
// Next, 256K allocated for the texture.
// Finally 4K allocated for the texture upload function to use for DMAing
// (The KILOBYTES macro just multiplies the number by 1024)
#define PRIMITIVEMEM_16K		0
#define TEXTUREMEM_256K		(KILOBYTES(16))
#define DMAMEM_4K			(KILOBYTES(272))
#define TOTALMEMNEEDED		(KILOBYTES(276))

// The Game mode...
#define	GAME_MODE_MENU		1
#define	GAME_MODE_PLAYING	2
#define	GAME_MODE_GAMEOVER	3

// The SPS2 handle.
int iSPS2Desc;

// Global variables.
int gameMode = 0;
int hiScore = 1968;
int Score = 0;
int Score1 = 0;
int Score2 = 0;
bool bPlayerOne = true;
bool bOnePlayerMode = true;
bool bGetReady = true;
bool bResetEnemy = true;
int counter = 0;
int spriteCount = 0;
int EnemyDir = 2;
int ExplosionDir = 2;
int shipPos = 200;
int maxX, minX;
int maxY, minY;
int tempX;
int tempY;
int i;
int bombStart = 3;
int shipsLeft = 2;
int shipsLeft1 = 2;
int shipsLeft2 = 2;
bool bShot = false;
bool bBomb = false;
bool bUFO = false;
bool bExtraShip = false;
bool bExplosion = false;
int explosionSpriteCounter = 0;

// The DMA registers, tags etc... we will be using.
Dn_CHCR_t chcr;
Dn_MADR_t madr;
sps2DMATag_t DMACnt;
sps2DMATag_t DMARef;
sps2DMATag_t DMAEnd;
uint64 * pDMATag;

// Memory handeling
sps2Memory_t * pMemCached;
char * pMemPosition;
char * pStartMemPosition;

// Number of QWords to transfer.
int iQWC = 0;

// Menu Sprites.
CSprite sprite1( 0, 0, 16, 16);
CSprite sprite2( 0, 32, 16, 16);
CSprite sprite3( 0, 64, 16, 16);
CSprite sprite4( 0, 192, 16, 16);
CSprite sprite5( 0, 160, 16, 16);
CSprite sprite6( 19, 166, 237, 90);

// Enemy sprites in game...
CSprite * spriteRow1[10];
CSprite * spriteRow2[10];
CSprite * spriteRow3[10];
CSprite * spriteRow4[10];
CSprite * spriteRow5[10];
CSprite * spriteRow6[10];

// Base Sprites
CSprite * spriteBase1Row1[16];
CSprite * spriteBase1Row2[16];
CSprite * spriteBase2Row1[16];
CSprite * spriteBase2Row2[16];
CSprite * spriteBase3Row1[16];
CSprite * spriteBase3Row2[16];

// Other sprites.
CSprite spriteShip( 0, 160, 16, 16);
CSprite spriteExtraShip1( 0, 160, 16, 16);
CSprite spriteExtraShip2( 0, 160, 16, 16);
CSprite spriteExtraShip3( 0, 160, 16, 16);
CSprite spriteUFO( 0, 192, 16, 16);
CSprite spriteShot( 16, 0, 2, 6);
CSprite spriteBomb( 16, 16, 2, 6);
CSprite spriteExplosion( 0, 96, 16, 16);

int main(void)
{

	// Create the enemy sprites...
	for (i = 0; i<10 ; i++) {
		spriteRow1[i] = new CSprite( 0, 0, 16, 16);
		spriteRow2[i] = new CSprite( 0, 0, 16, 16);
		spriteRow3[i] = new CSprite( 0, 32, 16, 16);
		spriteRow4[i] = new CSprite( 0, 32, 16, 16);
		spriteRow5[i] = new CSprite( 0, 64, 16, 16);
		spriteRow6[i] = new CSprite( 0, 64, 16, 16);
	}

	// Create the base sprites...
	for (i = 0; i<8 ; i++) {
		spriteBase1Row1[i] = new CSprite( i*2, 226, 2, 6);
		spriteBase1Row2[i] = new CSprite( i*2, 232, 2, 6);
		spriteBase1Row1[i+8] = new CSprite( i*2, 242, 2, 6);
		spriteBase1Row2[i+8] = new CSprite( i*2, 248, 2, 6);

		spriteBase2Row1[i] = new CSprite( i*2, 226, 2, 6);
		spriteBase2Row2[i] = new CSprite( i*2, 232, 2, 6);
		spriteBase2Row1[i+8] = new CSprite( i*2, 242, 2, 6);
		spriteBase2Row2[i+8] = new CSprite( i*2, 248, 2, 6);

		spriteBase3Row1[i] = new CSprite( i*2, 226, 2, 6);
		spriteBase3Row2[i] = new CSprite( i*2, 232, 2, 6);
		spriteBase3Row1[i+8] = new CSprite( i*2, 242, 2, 6);
		spriteBase3Row2[i+8] = new CSprite( i*2, 248, 2, 6);
	}

	// Initialize game mode and menu
	gameMode = GAME_MODE_MENU;
	MenuInit();

	// Initialise the Pad connected to port 1 of the PS2, we initilize it as analogue so we can read the pressure.
	if(pad_init (PAD_0, PAD_INIT_LOCK | PAD_INIT_ANALOGUE | PAD_INIT_PRESSURE ) == 0)
	{
		printf("Unable to initialise the pads\n");
		pad_cleanup (PAD_0);
		exit(0);
	}

	//initialise sps2
	iSPS2Desc = sps2Init();

	// Initialise the screen
	sps2UScreenInit(0);

	// Allocate the SPS2 memory
	pMemCached = sps2Allocate(TOTALMEMNEEDED, 
							SPS2_MAP_BLOCK_4K |	SPS2_MAP_CACHED, 
							iSPS2Desc);

	// Get a char pointer, we can use this to access memory addresses
	// relative to the start of SPS2 memory
	pStartMemPosition = (char *)pMemCached->pvStart;

	// The texture class for the game bitmap
	CTexture Tex(iSPS2Desc, &pStartMemPosition[TEXTUREMEM_256K], pMemCached);

	// Load game bitmap into SPS2 memory
	if(!Tex.LoadBitmap("space.bmp", true, false))
	{
		printf("Failed to load texture.");
		return -1;
	}
	
	//upload the game texture
	Tex.Upload(sps2UScreenGetFirstFreeGSPage(), &pStartMemPosition[DMAMEM_4K]);


// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- Main Loop -=*=-=*=-=*=-=*=-=*=-

	do
	{

		// Read the pad.
		pad_update (PAD_0);

		// Get first memory position from the allocated memory.
		pMemPosition = (char *) pMemCached->pvStart;

		// This is a general counter used for the menu etc...
		counter++;
		// Reset counter after 10 000 cycles, however the menu will reset this after 1000...
		if (counter > 10000) {
			counter = 0;
		}

		//clear the screen to the specified colour
		sps2UScreenClear(0x00, 0x00, 0x00);

		// Switch the game mode...
		switch (gameMode) {
		case GAME_MODE_MENU:
			GameMenu();
		break;
		case GAME_MODE_PLAYING:
			GamePlaying();
		break;
		case GAME_MODE_GAMEOVER:
			GameOver();
		break;
		default:
			GameMenu();
		break;
		}

		// Display the buffer we were just drawing to
		sps2UScreenSwap();

	// Press start and select together to end the loop, this can be done anytime in the game...
	}while(!((pad[0].buttons & PAD_START)&&(pad[0].buttons & PAD_SELECT)));
	
	// Shutdown the screen
	sps2UScreenShutdown();

	// Release SPS2
	sps2Release(iSPS2Desc);

	// Release pad.
	pad_cleanup (PAD_0);

	// Clean up enemy sprites
	for (i = 0; i<10 ; i++) {
		delete spriteRow1[i];
		delete spriteRow2[i];
		delete spriteRow3[i];
		delete spriteRow4[i];
		delete spriteRow5[i];
		delete spriteRow6[i];
	}

	// Clean upo the base sprites
	for (i = 0; i<16 ; i++) {
		delete spriteBase1Row1[i];
		delete spriteBase1Row2[i];
		delete spriteBase2Row1[i];
		delete spriteBase2Row2[i];
		delete spriteBase3Row1[i];
		delete spriteBase3Row2[i];
	}

	return 0;
}

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- End of program -=*=-=*=-=*=-=*=-=*=-


// Initialize the menu
void MenuInit(void){
	spriteCount = 6;
	counter = 0;
}


// Called once per frame for the menu
void GameMenu(void) {

	// Hide menu sprites.
	sprite1.setPos(-500,0);
	sprite2.setPos(-500,0);
	sprite3.setPos(-500,0);
	sprite4.setPos(-500,0);
	sprite5.setPos(-500,0);
	sprite6.setPos(160,55);

	// Check one player mode
	if(pad[0].buttons & PAD_LEFT) {
		bOnePlayerMode = true;		
	}

	// Check two player mode
	if(pad[0].buttons & PAD_RIGHT) {
		bOnePlayerMode = false;		
	}

	// Print text
	sps2UPrintf(0,"\377808080\376(50,10)SCORE <1>");
	sps2UPrintf(0,"\377808080\376(60,25)%d", Score1);

	sps2UPrintf(0,"\377808080\376(250,10)HI-SCORE");
	sps2UPrintf(0,"\377808080\376(260,25)%d", hiScore);

	sps2UPrintf(0,"\377808080\376(450,10)SCORE <2>");
	sps2UPrintf(0,"\377808080\376(460,25)%d", Score2);

	if (bOnePlayerMode) {
		sps2UPrintf(0,"\377808080\376(130,320)PRESS <RIGHT> TO SELECT TWO PLAYER MODE.");
	} else {
		sps2UPrintf(0,"\377808080\376(135,320)PRESS <LEFT> TO SELECT ONE PLAYER MODE.");
	}

	sps2UPrintf(0,"\377808080\376(210,340)PRESS <START> TO PLAY");
	sps2UPrintf(0,"\377808080\376(180,360)PRESS <START+SELECT> TO EXIT");


	if (bOnePlayerMode) {
		sps2UPrintf(0,"\377808080\376(230,175)ONE PLAYER");
	} else {
		sps2UPrintf(0,"\377808080\376(230,175)TWO PLAYERS");
	}

	// Flash GET READY after a while
	if (counter > 250) {
		if (counter%10>5) {
			sps2UPrintf(0,"\377FFFFFF\376(230,150)GET READY...");
		} else {
			sps2UPrintf(0,"\377808080\376(230,150)GET READY...");
		}
	}

	// Move the ship over the screen
	if (counter > 200 && counter < 400) {
		sprite5.setPos(counter, 400);
	}

	// Move the ship over the screen
	if (counter > 600 && counter < 800) {
		sprite5.setPos(1000-counter, 400);
	}

	// Display the 30 point enemy
	if (counter > 100) {
		sprite1.setPos(220,200);
		// Animate enemy
		sprite1.setTexturePos(0,(counter/10)%2);
		sps2UPrintf(0,"\377808080\376(240,200)= 30 POINTS.");
	}

	// Display the 20 point enemy
	if (counter > 300) {
		sprite2.setPos(220,220);
		// Animate enemy
		sprite2.setTexturePos(0,(counter/10)%2);
		sps2UPrintf(0,"\377808080\376(240,220)= 20 POINTS.");
	}

	// Display the 10 point enemy
	if (counter > 500) {
		sprite3.setPos(220,240);
		// Animate enemy
		sprite3.setTexturePos(0,(counter/10)%2);
		sps2UPrintf(0,"\377808080\376(240,240)= 10 POINTS.");
	}

	// Display the UFO
	if (counter > 700) {
		sprite4.setPos(220,255);
		// Animate enemy
		sprite4.setTexturePos(0,(counter/10)%2);
		sps2UPrintf(0,"\377808080\376(240,260)= ? POINTS.");
	}

	sps2UPrintfRender();

	// Reset counter after 1000 cycles
	if (counter > 1000) {
		counter = 0;
	}

	// Render Sprites...
	AddGIFTag(pMemPosition, spriteCount);
	sprite1.addSpriteToGIFTag(pMemPosition, 100);
	sprite2.addSpriteToGIFTag(pMemPosition, 100);
	sprite3.addSpriteToGIFTag(pMemPosition, 100);
	sprite4.addSpriteToGIFTag(pMemPosition, 100);
	sprite5.addSpriteToGIFTag(pMemPosition, 100);
	sprite6.addSpriteToGIFTag(pMemPosition, 100);

	// Configure the required register values
	// Set the address field to the physical address of the GS Packet
	madr.i32 = 0;		
	madr.s.ADDR = sps2GetPhysicalAddress((void *)&pStartMemPosition[PRIMITIVEMEM_16K], pMemCached);
	
	// The channel control register
	chcr.i32 = 0;
	chcr.s.MOD = CHCR_MOD_NORMAL;	// Normal mode transfer
	chcr.s.STR = 1;					// Start the transfer
	
	// Calculate Quad Words.
	iQWC = ((int)pMemPosition - (int)pStartMemPosition) >> 4;

	// DMA Channel 2 is the one we want, it goes to the GS
	// Amount of data to be transmitted
	*EE_D2_QWC = iQWC;

	// Set the memory address register
	*EE_D2_MADR = madr.i32;

	// Important: Make sure all the data is written to the physical memory
	sps2FlushCache(iSPS2Desc);	

	// Set the register to start the transfer
	*EE_D2_CHCR = chcr.i32;	

	// Wait for it to complete
	sps2WaitForDMA(2, iSPS2Desc);

	// Check to see if we started playing...
	if(pad[0].pressed & PAD_START) {
		GameInit();
		gameMode = 	GAME_MODE_PLAYING;
	}
	return;
}

// Init the game
void GameInit(void) {
	// Zero score
	Score = 0;
	Score1 = 0;
	Score2 = 0;
	bPlayerOne = true;
	bGetReady = true;
	shipsLeft = 2;
	shipsLeft1 = 2;
	if (bOnePlayerMode) {
		shipsLeft2 = -1;
	} else {
		shipsLeft2 = 2;
	}
	bResetEnemy = false;
	// Reset Enemies.
	for (i = 0; i<10 ; i++) {
		spriteRow1[i]->setPos(100+i*30, 100);
		spriteRow1[i]->setVisible(true);
		spriteRow2[i]->setPos(100+i*30, 120);
		spriteRow2[i]->setVisible(true);
		spriteRow3[i]->setPos(100+i*30, 140);
		spriteRow3[i]->setVisible(true);
		spriteRow4[i]->setPos(100+i*30, 160);
		spriteRow4[i]->setVisible(true);
		spriteRow5[i]->setPos(100+i*30, 180);
		spriteRow5[i]->setVisible(true);
		spriteRow6[i]->setPos(100+i*30, 200);
		spriteRow6[i]->setVisible(true);
	}
	// Reset Bases
	for (i = 0; i<16 ; i++) {
		spriteBase1Row1[i]->setPos(150+i*2, 330);
		spriteBase1Row1[i]->setVisible(true);
		spriteBase1Row2[i]->setPos(150+i*2, 336);
		spriteBase1Row2[i]->setVisible(true);

		spriteBase2Row1[i]->setPos(304+i*2, 330);
		spriteBase2Row1[i]->setVisible(true);
		spriteBase2Row2[i]->setPos(304+i*2, 336);
		spriteBase2Row2[i]->setVisible(true);

		spriteBase3Row1[i]->setPos(490+i*2, 330);
		spriteBase3Row1[i]->setVisible(true);
		spriteBase3Row2[i]->setPos(490+i*2, 336);
		spriteBase3Row2[i]->setVisible(true);
	}

	spriteCount = 0;
	EnemyDir = 1;
	shipPos = 220;
	spriteShip.setPos(shipPos,350);
	spriteExtraShip1.setPos(100,380);
	spriteExtraShip2.setPos(130,380);
	spriteExtraShip3.setPos(-100,380);
	spriteUFO.setPos(-20,70);
	spriteShot.setPos(-20,0);
	spriteBomb.setPos(-20,0);
	spriteExplosion.setPos(-20,0);
	shipsLeft = 2;
	bExplosion = false;
	bShot = false;
	bBomb = false;
	bUFO = false;
	bExtraShip = false;
	maxX = 370;
	minX = 100;
	counter = 1;
	explosionSpriteCounter = 0;
}

void GamePlaying(void) {

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- Start Print Text -=*=-=*=-=*=-=*=-=*=-

	// Print static text...
	sps2UPrintf(0,"\377808080\376(50,10)SCORE <1>");
	sps2UPrintf(0,"\377808080\376(60,25)%d", Score1);

	sps2UPrintf(0,"\377808080\376(250,10)HI-SCORE");
	sps2UPrintf(0,"\377808080\376(260,25)%d", hiScore);

	sps2UPrintf(0,"\377808080\376(450,10)SCORE <2>");
	sps2UPrintf(0,"\377808080\376(460,25)%d", Score2);

	// Print get ready text..
	if (bGetReady) {
		if (bPlayerOne) {
			// Nice flashing effect...
			if (counter%10>5) {
				sps2UPrintf(0,"\377FFFFFF\376(200,50)PLAYER ONE GET READY...");
			} else {
				sps2UPrintf(0,"\377808080\376(200,50)PLAYER ONE GET READY...");
			}
		} else {
			if (counter%10>5) {
				sps2UPrintf(0,"\377FFFFFF\376(200,50)PLAYER TWO GET READY...");
			} else {
				sps2UPrintf(0,"\377808080\376(200,50)PLAYER TWO GET READY...");
			}
		}
		// If we are in Get Ready mode, start the game after a while
		if (counter>200) {
			bGetReady = false;
			counter = 1;
		}		
	}


	sps2UPrintfRender();

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- End Print Text -=*=-=*=-=*=-=*=-=*=-

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- Start Inpuit from Pad -=*=-=*=-=*=-=*=-=*=-

	// Input from Pad, move ship left right or shoot...

	// Wait for get ready to end...
	if (!bGetReady) {
		// Update Ship position. We move two pixels at the time.
		if (pad[0].buttons & PAD_LEFT) {
			if (shipPos>40) {
				shipPos--;
				shipPos--;
			}
		}
		if (pad[0].buttons & PAD_RIGHT) {
			if (shipPos<600) {
				shipPos++;
				shipPos++;
			}
			if (bGetReady) {
				bGetReady = false;
			}
		}
	}
	spriteShip.setPos(shipPos,350);

	// Shooot.
	if(pad[0].pressed & PAD_CROSS) {
		// Only shoot if we are playing and there is not ready a shot fired.
		if (!bShot & !bGetReady) {
			bShot = true;
			spriteShot.setPos(shipPos+7,344);
		}
		// If we are not playing, start the game if a small amount of time has elapsed.
		// Don't shoot, if you want to shoot, move this above "if (!bShot & !bGetReady) {"
		if (bGetReady & counter > 50) {
			bGetReady = false;
			counter = 1;
		}
	}

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- End Inpuit from Pad -=*=-=*=-=*=-=*=-=*=-

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- Start Animation -=*=-=*=-=*=-=*=-=*=-
// Animate stuff, either move or set a different texture.


// Don't animate unless the game has begun...
if (!bGetReady) {

	// Animate the "shooting" effect.
	if (spriteShot.getYPos() > 330) {
		spriteShip.setTexturePos(0,1);
	} else {
		spriteShip.setTexturePos(0,0);
	}

	// Animate, move and check sprites for max x and min x.
	maxX = 0; minX = 600, maxY = 0;
	for (i = 0; i<10 ; i++) {
		// Top Row.
		// Animate enemy.
		spriteRow1[i]->setTexturePos(0,(counter/20)%2);
		// Move enemy
		spriteRow1[i]->movePos(EnemyDir, 0);
		// If enemy is visible, calculate max and min values
		if (spriteRow1[i]->getVisible()) {
			tempX = spriteRow1[i]->getXPos();
			tempY = spriteRow1[i]->getYPos();
			maxY = (maxY > tempY) ? maxY : tempY;
			maxX = (maxX > tempX) ? maxX : tempX;
			minX = (minX < tempX) ? minX : tempX;
		}

		// Second Row.
		spriteRow2[i]->setTexturePos(0,(counter/20+1)%2);
		spriteRow2[i]->movePos(EnemyDir, 0);
		if (spriteRow2[i]->getVisible()) {
			tempX = spriteRow2[i]->getXPos();
			tempY = spriteRow2[i]->getYPos();
			maxY = (maxY > tempY) ? maxY : tempY;
			maxX = (maxX > tempX) ? maxX : tempX;
			minX = (minX < tempX) ? minX : tempX;
		}

		// Third Row.
		spriteRow3[i]->setTexturePos(0,(counter/20)%2);
		spriteRow3[i]->movePos(EnemyDir, 0);
		if (spriteRow3[i]->getVisible()) {
			tempX = spriteRow3[i]->getXPos();
			tempY = spriteRow3[i]->getYPos();
			maxY = (maxY > tempY) ? maxY : tempY;
			maxX = (maxX > tempX) ? maxX : tempX;
			minX = (minX < tempX) ? minX : tempX;
		}

		// Fourth Row.
		spriteRow4[i]->setTexturePos(0,(counter/20+1)%2);
		spriteRow4[i]->movePos(EnemyDir, 0);
		if (spriteRow4[i]->getVisible()) {
			tempX = spriteRow4[i]->getXPos();
			tempY = spriteRow4[i]->getYPos();
			maxY = (maxY > tempY) ? maxY : tempY;
			maxX = (maxX > tempX) ? maxX : tempX;
			minX = (minX < tempX) ? minX : tempX;
		}

		// Fifth Row.
		spriteRow5[i]->setTexturePos(0,(counter/20)%2);
		spriteRow5[i]->movePos(EnemyDir, 0);
		if (spriteRow5[i]->getVisible()) {
			tempX = spriteRow5[i]->getXPos();
			tempY = spriteRow5[i]->getYPos();
			maxY = (maxY > tempY) ? maxY : tempY;
			maxX = (maxX > tempX) ? maxX : tempX;
			minX = (minX < tempX) ? minX : tempX;
		}

		// Sixth Row.
		spriteRow6[i]->setTexturePos(0,(counter/20+1)%2);
		spriteRow6[i]->movePos(EnemyDir, 0);
		if (spriteRow6[i]->getVisible()) {
			tempX = spriteRow6[i]->getXPos();
			tempY = spriteRow6[i]->getYPos();
			maxY = (maxY > tempY) ? maxY : tempY;
			maxX = (maxX > tempX) ? maxX : tempX;
			minX = (minX < tempX) ? minX : tempX;
		}

	}

	// Animate shot...
	if (bShot) {
		spriteShot.movePos(0, -2);
	}
	// Test if we are done with the shot.
	if (spriteShot.getYPos() < 50) {
		spriteShot.setPos(-500,0);
		bShot = false;
	}

	// Animate bomb...
	if (bBomb) {
		spriteBomb.movePos(0, 2);
	}
	// Test if we are done with the bomb.
	if (spriteBomb.getYPos() > 370) {
		spriteBomb.setPos(-500,0);
		bBomb = false;
	}

	// Animate UFO.
	if (bUFO) {
		spriteUFO.movePos(1, 0);
		spriteUFO.setTexturePos(0,(counter/10)%2);
	}

	// Animate explosion.
	if ( bExplosion ) {
		// Slow explosion down, animate every third frame
		spriteExplosion.setTexturePos(0,explosionSpriteCounter/3);
		// Move with enemies or not at all for UFO.
		spriteExplosion.movePos(ExplosionDir, 0);
		// Animate and test if explosion is done
		explosionSpriteCounter++;
		if (explosionSpriteCounter>12) {
			bExplosion = false;
			spriteExplosion.setPos(-100,0);
			explosionSpriteCounter = 0;
		}
	}

}

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- END Animation -=*=-=*=-=*=-=*=-=*=-

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- Start Testing  Stuff -=*=-=*=-=*=-=*=-=*=-

	// Test to see if we need to change direction of the enemies...
	if (EnemyDir > 0) {
		if (maxX>600) {
			EnemyDir = -EnemyDir;
			// Move enemies down a little bit.
			for (i = 0; i<10 ; i++) {
				spriteRow1[i]->movePos(0, 6);
				spriteRow2[i]->movePos(0, 6);
				spriteRow3[i]->movePos(0, 6);
				spriteRow4[i]->movePos(0, 6);
				spriteRow5[i]->movePos(0, 6);
				spriteRow6[i]->movePos(0, 6);
			}
		}
	}

	// Test to see if we need to change direction of the enemies...
	if (EnemyDir < 0) {
		if (minX<40) {
			EnemyDir = -EnemyDir;
			// Move enemies down a little bit.
			for (i = 0; i<10 ; i++) {
				spriteRow1[i]->movePos(0, 6);
				spriteRow2[i]->movePos(0, 6);
				spriteRow3[i]->movePos(0, 6);
				spriteRow4[i]->movePos(0, 6);
				spriteRow5[i]->movePos(0, 6);
				spriteRow6[i]->movePos(0, 6);
			}
		}
	}


	// Did the UFO pass
	if (spriteUFO.getXPos() > 670) {
		spriteUFO.setPos(-20,70);
		bUFO = false;
	}

	// Did we hit the UFO ???
	if (spriteUFO.inside(spriteShot)) {
		// Random score between 50 and 100.
		Score = Score + (spriteUFO.getXPos()%50) + 50;
		// Start explosion, set position and speed.
		bExplosion = true;
		ExplosionDir = 1;
		spriteExplosion.setPos(spriteUFO.getXPos(),spriteUFO.getYPos());
		// Hide amd stop UFO and shot.
		spriteUFO.setPos(-20,70);
		spriteShot.setPos(-500,0);
		bShot = false;
		bUFO = false;
	}

	// Show UFO every 2000 counter.
	if (!bUFO && (counter%2000 == 0)) {
		bUFO = true;
	}

	// Did we hit the Base ???
	for (i = 0; i<16 ; i++) {
		// Test shot, top left base.
		if (spriteBase1Row1[i]->inside(spriteShot)) {
			spriteBase1Row1[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}
		// Test bomb, top left base.
		if (spriteBase1Row1[i]->inside(spriteBomb)) {
			spriteBase1Row1[i]->setVisible(false);
			spriteBomb.setPos(-100,0);
			bBomb = false;
		}
		// Test shot, bottom left base.
		if (spriteBase1Row2[i]->inside(spriteShot)) {
			spriteBase1Row2[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}
		// Test bomb, bottom left base.
		if (spriteBase1Row2[i]->inside(spriteBomb)) {
			spriteBase1Row2[i]->setVisible(false);
			spriteBomb.setPos(-100,0);
			bBomb = false;
		}

		// Test shot, top middle base.
		if (spriteBase2Row1[i]->inside(spriteShot)) {
			spriteBase2Row1[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}
		// Test bomb, top middle base.
		if (spriteBase2Row1[i]->inside(spriteBomb)) {
			spriteBase2Row1[i]->setVisible(false);
			spriteBomb.setPos(-100,0);
			bBomb = false;
		}
		// Test shot, bottom middle base.
		if (spriteBase2Row2[i]->inside(spriteShot)) {
			spriteBase2Row2[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}
		// Test bomb, bottom middle base.
		if (spriteBase2Row2[i]->inside(spriteBomb)) {
			spriteBase2Row2[i]->setVisible(false);
			spriteBomb.setPos(-100,0);
			bBomb = false;
		}

		// Test shot, top right base.
		if (spriteBase3Row1[i]->inside(spriteShot)) {
			spriteBase3Row1[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}
		// Test bomb, top right base.
		if (spriteBase3Row1[i]->inside(spriteBomb)) {
			spriteBase3Row1[i]->setVisible(false);
			spriteBomb.setPos(-100,0);
			bBomb = false;
		}
		// Test shot, bottom right base.
		if (spriteBase3Row2[i]->inside(spriteShot)) {
			spriteBase3Row2[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}
		// Test bomb, bottom right base.
		if (spriteBase3Row2[i]->inside(spriteBomb)) {
			spriteBase3Row2[i]->setVisible(false);
			spriteBomb.setPos(-100,0);
			bBomb = false;
		}
	}

	// Test to see if we hit an enemy...	
	for (i = 0; i<10 ; i++) {
		// Top row.
		if (spriteRow1[i]->inside(spriteShot)) {
			// Increase score.
			Score = Score + 30;
			// Start explosion and set sped and position.
			bExplosion = true;
			ExplosionDir = EnemyDir;
			spriteExplosion.setPos(spriteRow1[i]->getXPos(),spriteRow1[i]->getYPos());
			// Hide enemy.
			spriteRow1[i]->setVisible(false);
			// End and hide shot.
			spriteShot.setPos(-500,0);
			bShot = false;
		}

		// Second row.
		if (spriteRow2[i]->inside(spriteShot)) {
			Score = Score + 30;
			bExplosion = true;
			ExplosionDir = EnemyDir;
			spriteExplosion.setPos(spriteRow2[i]->getXPos(),spriteRow2[i]->getYPos());
			spriteRow2[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}

		// Third Row.
		if (spriteRow3[i]->inside(spriteShot)) {
			Score = Score + 20;
			bExplosion = true;
			ExplosionDir = EnemyDir;
			spriteExplosion.setPos(spriteRow3[i]->getXPos(),spriteRow3[i]->getYPos());
			spriteRow3[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}

		// Fourth row.
		if (spriteRow4[i]->inside(spriteShot)) {
			Score = Score + 20;
			bExplosion = true;
			ExplosionDir = EnemyDir;
			spriteExplosion.setPos(spriteRow4[i]->getXPos(),spriteRow4[i]->getYPos());
			spriteRow4[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}

		// Fifth row.
		if (spriteRow5[i]->inside(spriteShot)) {
			Score = Score + 10;
			bExplosion = true;
			ExplosionDir = EnemyDir;
			spriteExplosion.setPos(spriteRow5[i]->getXPos(),spriteRow5[i]->getYPos());
			spriteRow5[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}

		// Bottom row.
		if (spriteRow6[i]->inside(spriteShot)) {
			Score = Score + 10;
			bExplosion = true;
			ExplosionDir = EnemyDir;
			spriteExplosion.setPos(spriteRow6[i]->getXPos(),spriteRow6[i]->getYPos());
			spriteRow6[i]->setVisible(false);
			spriteShot.setPos(-500,0);
			bShot = false;
		}
	}

	// Count Enemy sprites...
	spriteCount = 0;
	for (i = 0; i<10 ; i++) {
		if (spriteRow1[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteRow2[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteRow3[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteRow4[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteRow5[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteRow6[i]->getVisible()) {
			spriteCount++;
		}
	}

	// If we hit the bomb with the shot, remove it...
	if (spriteBomb.inside(spriteShot))  {
		spriteBomb.setPos(-100,0);
		spriteShot.setPos(-100,0);
		bShot = false;
		bBomb = false;
	}

	// Bonus ship @ 5000 points.
	if (!bExtraShip & Score >= 5000) {
		bExtraShip = true;
		shipsLeft++;
	}

	// Set score for active player.
	if (bPlayerOne) {
		Score1 = Score;
	} else {
		Score2 = Score;
	}

	// Did we get a new highscore.
	if (Score > hiScore) {
		hiScore = Score;
	}

	// Did we got hit or did the Enemy land...
	if (spriteBomb.inside(spriteShip) | (maxY > 340))  {
		shipsLeft--;
		// Set correct ships left...
		if (bPlayerOne) {
			shipsLeft1 = shipsLeft;
		} else {
			shipsLeft2 = shipsLeft;
		}
		spriteCount = 0;
		spriteBomb.setPos(-100,0);
		spriteShot.setPos(-500,0);
		spriteExplosion.setPos(-20,0);
		bShot = false;
		bBomb = false;
		// Handle one or two player mode...
		if (!bOnePlayerMode) {
			if (bPlayerOne & shipsLeft2 > -1) {
				bPlayerOne = false;
				shipsLeft = shipsLeft2;
				Score = Score2;
			} else {
				bPlayerOne = true;
				shipsLeft = shipsLeft1;
				Score = Score1;
			}
		}
		// Reset game...
		bGetReady = true;
		bResetEnemy = true;
		counter = 1;
		// Test to see if it is Game Over.
		if ((shipsLeft1+shipsLeft2) == -2) {
			GameOverInit();
			gameMode = 	GAME_MODE_GAMEOVER;
			return;
		}
	}

	// If all enemies have been shot, start all over again.
	if (spriteCount == 0) {
		for (i = 0; i<10 ; i++) {
			spriteRow1[i]->setPos(100+i*30, 100);
			spriteRow1[i]->setVisible(true);
			spriteRow2[i]->setPos(100+i*30, 120);
			spriteRow2[i]->setVisible(true);
			spriteRow3[i]->setPos(100+i*30, 140);
			spriteRow3[i]->setVisible(true);
			spriteRow4[i]->setPos(100+i*30, 160);
			spriteRow4[i]->setVisible(true);
			spriteRow5[i]->setPos(100+i*30, 180);
			spriteRow5[i]->setVisible(true);
			spriteRow6[i]->setPos(100+i*30, 200);
			spriteRow6[i]->setVisible(true);
		}
		spriteCount = 60;
		// Are we just reseting the game, or did we actually shoot all of them???
		if (bResetEnemy) {
			bResetEnemy = false;
		} else {
			// Bonus score for shooting all enemies.
			Score = Score + (counter%50) + 50;
		}
		// Remove Bomb and shot.
		spriteBomb.setPos(-100,0);
		spriteShot.setPos(-100,0);
		bShot = false;
		bBomb = false;
		// Reset Bases
		for (i = 0; i<16 ; i++) {
			spriteBase1Row1[i]->setPos(150+i*2, 330);
			spriteBase1Row1[i]->setVisible(true);
			spriteBase1Row2[i]->setPos(150+i*2, 336);
			spriteBase1Row2[i]->setVisible(true);
		
			spriteBase2Row1[i]->setPos(304+i*2, 330);
			spriteBase2Row1[i]->setVisible(true);
			spriteBase2Row2[i]->setPos(304+i*2, 336);
			spriteBase2Row2[i]->setVisible(true);

			spriteBase3Row1[i]->setPos(490+i*2, 330);
			spriteBase3Row1[i]->setVisible(true);
			spriteBase3Row2[i]->setPos(490+i*2, 336);
			spriteBase3Row2[i]->setVisible(true);
		}
	}

	// Draw the extra ship sprites in the bottom of the screen.
	if (shipsLeft > 0) {
		spriteExtraShip1.setPos(100,380);
	} else {
		spriteExtraShip1.setPos(-100,380);
	}
	if (shipsLeft > 1) {
		spriteExtraShip2.setPos(130,380);
	} else {
		spriteExtraShip2.setPos(-130,380);
	}
	if (shipsLeft > 2) {
		spriteExtraShip3.setPos(160,380);
	} else {
		spriteExtraShip3.setPos(-160,380);
	}

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- End Testing  Stuff -=*=-=*=-=*=-=*=-=*=-

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- Start Calculate Stuff -=*=-=*=-=*=-=*=-=*=-


	// Calculate speed depending on howe many sprites we have on screen-
	i=1;
	if (spriteCount<20) {
		i=2;
	}
	if (spriteCount<5) {
		i=3;
	}
	// Make sure we go in the right direction
	if (EnemyDir>0) {
		EnemyDir = i;
	} else {
		EnemyDir = -i;
	}

	// Bombs away...
	// Start with top row, if there are any visible ships, make it drop a bomb...
	// Then just work down to the bottom row testing to see if the ship is visible.
	// If it is, we'll move the bomb to the next row.
	// If no ships are available move to next column.
	// Then move the next enemy to drop a bomb three columns to the right.
	if (!bBomb) {
		for (i = bombStart; i<(10+bombStart) & !bBomb ; i++) {
			if (spriteRow1[i%10]->getVisible()) {
				bBomb = true;
				spriteBomb.setPos(spriteRow1[i%10]->getXPos()+7,spriteRow1[i%10]->getYPos()+15);
			}
			if (spriteRow2[i%10]->getVisible()) {
				bBomb = true;
				spriteBomb.setPos(spriteRow2[i%10]->getXPos()+7,spriteRow2[i%10]->getYPos()+15);
			}
			if (spriteRow3[i%10]->getVisible()) {
				bBomb = true;
				spriteBomb.setPos(spriteRow3[i%10]->getXPos()+7,spriteRow3[i%10]->getYPos()+15);
			}
			if (spriteRow4[i%10]->getVisible()) {
				bBomb = true;
				spriteBomb.setPos(spriteRow4[i%10]->getXPos()+7,spriteRow4[i%10]->getYPos()+15);
			}
			if (spriteRow5[i%10]->getVisible()) {
				bBomb = true;
				spriteBomb.setPos(spriteRow5[i%10]->getXPos()+7,spriteRow5[i%10]->getYPos()+15);
			}
			if (spriteRow6[i%10]->getVisible()) {
				bBomb = true;
				spriteBomb.setPos(spriteRow6[i%10]->getXPos()+7,spriteRow6[i%10]->getYPos()+15);
			}
		}
		bombStart++;
		bombStart++;
		bombStart++;
	}	

// -=*=-=*=-=*=-=*=-=*=-=*=-=*=- End Calculate Stuff -=*=-=*=-=*=-=*=-=*=-

// -=*=-=*=-=*=-=*=-=*=-=*=- Start Render Sprites -=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-


	// Count Enemy sprites...
	spriteCount = 0;
	for (i = 0; i<10 ; i++) {
		if (spriteRow1[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteRow2[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteRow3[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteRow4[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteRow5[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteRow6[i]->getVisible()) {
			spriteCount++;
		}
	}

	// Add the Ship, Explosion, UFO, Bomb and shot.
	spriteCount++;
	spriteCount++;
	spriteCount++;
	spriteCount++;
	spriteCount++;

	// Add Extra ships.
	spriteCount++;
	spriteCount++;
	spriteCount++;

	// Count Base sprites...
	for (i = 0; i<16 ; i++) {
		if (spriteBase1Row1[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteBase1Row2[i]->getVisible()) {
			spriteCount++;
		}

		if (spriteBase2Row1[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteBase2Row2[i]->getVisible()) {
			spriteCount++;
		}

		if (spriteBase3Row1[i]->getVisible()) {
			spriteCount++;
		}
		if (spriteBase3Row2[i]->getVisible()) {
			spriteCount++;
		}
	}

	// Render Sprites...
	AddGIFTag(pMemPosition, spriteCount);
	// Render Enemies
	for (i = 0; i<10 ; i++) {
		spriteRow1[i]->addSpriteToGIFTag(pMemPosition, 100);
		spriteRow2[i]->addSpriteToGIFTag(pMemPosition, 100);
		spriteRow3[i]->addSpriteToGIFTag(pMemPosition, 100);
		spriteRow4[i]->addSpriteToGIFTag(pMemPosition, 100);
		spriteRow5[i]->addSpriteToGIFTag(pMemPosition, 100);
		spriteRow6[i]->addSpriteToGIFTag(pMemPosition, 100);
	}
	// Render Ship.
	spriteShip.addSpriteToGIFTag(pMemPosition, 100);
	// Render Shot.
	spriteShot.addSpriteToGIFTag(pMemPosition, 200);
	// Render Bomb.
	spriteBomb.addSpriteToGIFTag(pMemPosition, 200);
	// Render UFO.
	spriteUFO.addSpriteToGIFTag(pMemPosition, 100);
	// Render Explosion.
	spriteExplosion.addSpriteToGIFTag(pMemPosition, 100);
	// Render extra ships.
	spriteExtraShip1.addSpriteToGIFTag(pMemPosition, 100);
	spriteExtraShip2.addSpriteToGIFTag(pMemPosition, 100);
	spriteExtraShip3.addSpriteToGIFTag(pMemPosition, 100);

	// Render Base.
	for (i = 0; i<16 ; i++) {
		spriteBase1Row1[i]->addSpriteToGIFTag(pMemPosition, 50);
		spriteBase1Row2[i]->addSpriteToGIFTag(pMemPosition, 50);
		spriteBase2Row1[i]->addSpriteToGIFTag(pMemPosition, 50);
		spriteBase2Row2[i]->addSpriteToGIFTag(pMemPosition, 50);
		spriteBase3Row1[i]->addSpriteToGIFTag(pMemPosition, 50);
		spriteBase3Row2[i]->addSpriteToGIFTag(pMemPosition, 50);
	}


	// Calculate Quad Words.
	iQWC = ((int)pMemPosition - (int)pStartMemPosition) >> 4;

	// Set the pDMATag to start of DMA memory.
	pDMATag = (uint64 *)(&pStartMemPosition[DMAMEM_4K]);

	i = 0;
	do {
		DMARef.i64 = 0;
		if (iQWC>256){
			DMARef.s.QWC = 256;	// Transfer no more than 4K at a time
			iQWC = iQWC - 256;
		} else {
			DMARef.s.QWC = iQWC; // Transfer the leftover QWC.
			iQWC = 0;
		}
		DMARef.s.ID = DMA_ID_SOURCE_REF;
		// Get the physical address of each 4K page, and put it into the ref tag.
		DMARef.s.ADDR = sps2GetPhysicalAddress((char *)&pStartMemPosition[PRIMITIVEMEM_16K] + 4096*i, pMemCached);
		// Add the ref tag to the DMA chain.
		*pDMATag = DMARef.i64;
		pDMATag += 2;
		i++;
	} while (iQWC > 0);

	
	// Now that we have uploaded the image, the DMA chain needs to be ended. We will use
	// this final packet to set a few GS registers that will be used for texturing.

	DMAEnd.i64 = 0;
	DMAEnd.s.QWC = 0;	
	DMAEnd.s.ID = DMA_ID_SOURCE_END;

	// Add the DMA tag to the chain
	*pDMATag = DMAEnd.i64;
	pDMATag += 2;

	// Flush the DMA chain to memory before we fire it off
	sps2FlushCache(iSPS2Desc);

	// Don't transfer any data from MADR
	*EE_D2_QWC = 0;

	// Set the address of the first tag (right at the start of the chain)
	madr.i32 = 0;
	madr.s.ADDR = sps2GetPhysicalAddress(&pStartMemPosition[DMAMEM_4K], pMemCached);
	*EE_D2_TADR = madr.i32;

	// Set the channel control register, we are using chain mode.
	chcr.i32 = 0;
	chcr.s.MOD = CHCR_MOD_CHAIN;
	chcr.s.STR = 1;

	// Important: Make sure all the data is written to the physical memory
	sps2FlushCache(iSPS2Desc);	

	// Set the register to start the transfer
	*EE_D2_CHCR = chcr.i32;	

	// Wait for it to complete
	sps2WaitForDMA(2, iSPS2Desc);

// -=*=-=*=-=*=-=*=-=*=-=*=- End Render Sprites -=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-

	if (pad[0].pressed & PAD_TRI)   {
		GameOverInit();
		gameMode = 	GAME_MODE_GAMEOVER;
	}
	return;
}

void GameOverInit(void) {
}

void GameOver(void) {

	// Static text.
	sps2UPrintf(0,"\377808080\376(190,20)PRESS SQUARE TO CONTINUE.");

	// Print GAME OVER sequantally.
	if (counter <= 200) {
		if (counter > 20) {
			sps2UPrintf(0,"\377808080\376(240,200)G");
		}
		if (counter > 40) {
			sps2UPrintf(0,"\377808080\376(240,200)GA");
		}
		if (counter > 60) {
			sps2UPrintf(0,"\377808080\376(240,200)GAM");
		}
		if (counter > 80) {
			sps2UPrintf(0,"\377808080\376(240,200)GAME");
		}
		if (counter > 100) {
			sps2UPrintf(0,"\377808080\376(240,200)GAME O");
		}
		if (counter > 120) {
			sps2UPrintf(0,"\377808080\376(240,200)GAME OV");
		}
		if (counter > 140) {
			sps2UPrintf(0,"\377808080\376(240,200)GAME OVE");
		}
		if (counter > 160) {
			sps2UPrintf(0,"\377808080\376(240,200)GAME OVER");
		}
	}
	// Start flasing GAME OVER after a while...
	if (counter > 200) {
		if (counter%10>5) {
			sps2UPrintf(0,"\377FFFFFF\376(240,200)GAME OVER");
		} else {
			sps2UPrintf(0,"\377808080\376(240,200)GAME OVER");
		}
	}

	// Render it.
	sps2UPrintfRender();

	// Test to see if a certain time has elepsed or the user pressed the Square.
	if ((pad[0].pressed & PAD_SQUARE)| counter > 1000) {
		MenuInit();
		gameMode = 	GAME_MODE_MENU;
	}
	return;
}

// Just add the GIF Tag to the beginning of the allocated Memory.
void AddGIFTag(char *& pMem, int numSprites) {

		// Add a GIF Tag 
		sps2GIFTag_t * pGIFTag = (sps2GIFTag_t *) pMem;
 
	  	// Create the GIFTag  
  		pGIFTag->i128=0;       			// Blank it out.
	  	pGIFTag->s.NLOOP=numSprites;   	// How many sprites we are rendering.
  		pGIFTag->s.PRE=1;      			// Send our PRIM value to the GIF via this GIFTag.
	  	pGIFTag->s.EOP=1;      			// This is the end of the packets, no more GIFTags are to be expected after the sprites.
		pGIFTag->s.PRIM = GS_PRIM(PRIM_SPRITE, 
							PRIM_IIP_FLAT, 
							PRIM_TME_ON,
							PRIM_FGE_OFF, 
							PRIM_ABE_ON, 
							PRIM_AA1_OFF, 
							PRIM_FST_UV, 
							PRIM_CTXT_CONTEXT1, 
							PRIM_FIX_NOFIXDDA);
	  	pGIFTag->s.FLG=GIF_FLG_PACKED;	// Use packed format for data.
  		pGIFTag->s.NREG=0;     			// Start at 0 and increment as needed.
  
	  	// Add RGBAQ register.
  		SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_RGBAQ);
	  	// Add UV register.
  		SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_UV);
	  	// Add XYZ2 register.
  		SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_XYZ2);
	  	// Add UV register.
  		SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_UV);
	  	// Add XYZ2 register.
  		SPS2_SET_GIF_REG(*pGIFTag, pGIFTag->s.NREG++, GIF_REG_XYZ2);
		pGIFTag++;

		// Update memory pointer.
		pMem = (char *)pGIFTag;
}

